# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import base64
import pickle

import requests

from quasarr.providers.log import info


def create_and_persist_session(shared_state):
    nx = shared_state.values["config"]("Hostnames").get("nx")

    nx_session = requests.Session()

    cookies = {}
    headers = {
        'User-Agent': shared_state.values["user_agent"],
    }

    json_data = {
        'username': shared_state.values["config"]("NX").get("user"),
        'password': shared_state.values["config"]("NX").get("password")
    }

    nx_response = nx_session.post(f'https://{nx}/api/user/auth', cookies=cookies, headers=headers, json=json_data,
                                  timeout=10)

    error = False
    if nx_response.status_code == 200:
        try:
            response_data = nx_response.json()
            if response_data.get('err', {}).get('status') == 403:
                info("Invalid NX credentials provided.")
                error = True
            elif response_data.get('user').get('username') != shared_state.values["config"]("NX").get("user"):
                info("Invalid NX response on login.")
                error = True
            else:
                sessiontoken = response_data.get('user').get('sessiontoken')
                nx_session.cookies.set('sessiontoken', sessiontoken, domain=nx)
        except ValueError:
            info("Could not parse NX response on login.")
            error = True

        if error:
            shared_state.values["config"]("NX").save("user", "")
            shared_state.values["config"]("NX").save("password", "")
            return None

        serialized_session = pickle.dumps(nx_session)
        session_string = base64.b64encode(serialized_session).decode('utf-8')
        shared_state.values["database"]("sessions").update_store("nx", session_string)
        return nx_session
    else:
        info("Could not create NX session")
        return None


def retrieve_and_validate_session(shared_state):
    session_string = shared_state.values["database"]("sessions").retrieve("nx")
    if not session_string:
        nx_session = create_and_persist_session(shared_state)
    else:
        try:
            serialized_session = base64.b64decode(session_string.encode('utf-8'))
            nx_session = pickle.loads(serialized_session)
            if not isinstance(nx_session, requests.Session):
                raise ValueError("Retrieved object is not a valid requests.Session instance.")
        except Exception as e:
            info(f"Session retrieval failed: {e}")
            nx_session = create_and_persist_session(shared_state)

    return nx_session



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re

import requests
from bs4 import BeautifulSoup

from quasarr.providers.log import info
from quasarr.providers.sessions.nx import retrieve_and_validate_session


def get_filer_folder_links_via_api(shared_state, url):
    try:
        headers = {
            'User-Agent': shared_state.values["user_agent"],
            'Referer': url
        }

        m = re.search(r"/folder/([A-Za-z0-9]+)", url)
        if not m:
            return url  # not a folder URL

        folder_hash = m.group(1)
        api_url = f"https://filer.net/api/folder/{folder_hash}"

        response = requests.get(api_url, headers=headers, timeout=10)
        if not response or response.status_code != 200:
            return url

        data = response.json()
        files = data.get("files", [])
        links = []

        # Build download URLs from their file hashes
        for f in files:
            file_hash = f.get("hash")
            if not file_hash:
                continue
            dl_url = f"https://filer.net/get/{file_hash}"
            links.append(dl_url)

        # Return extracted links or fallback
        return links if links else url

    except:
        return url


def get_nx_download_links(shared_state, url, mirror, title): # signature must align with other download link functions!
    nx = shared_state.values["config"]("Hostnames").get("nx")

    if f"{nx}/release/" not in url:
        info("Link is not a Release link, could not proceed:" + url)

    nx_session = retrieve_and_validate_session(shared_state)
    if not nx_session:
        info(f"Could not retrieve valid session for {nx}")
        return []

    headers = {
        'User-Agent': shared_state.values["user_agent"],
        'Referer': url
    }

    json_data = {}

    url_segments = url.split('/')
    payload_url = '/'.join(url_segments[:-2]) + '/api/getLinks/' + url_segments[-1]

    payload = nx_session.post(payload_url,
                              headers=headers,
                              json=json_data,
                              timeout=10
                              )

    if payload.status_code == 200:
        try:
            payload = payload.json()
        except:
            info("Invalid response decrypting " + str(title) + " URL: " + str(url))
            shared_state.values["database"]("sessions").delete("nx")
            return []

    if payload and any(key in payload for key in ("err", "error")):
        error_msg = payload.get("err") or payload.get("error")
        info(f"Error decrypting {title!r} URL: {url!r} - {error_msg}")
        shared_state.values["database"]("sessions").delete("nx")
        return []

    try:
        decrypted_url = payload['link'][0]['url']
        if decrypted_url:
            if "filer.net/folder/" in decrypted_url:
                urls = get_filer_folder_links_via_api(shared_state, decrypted_url)
            else:
                urls = [decrypted_url]
            return urls
    except:
        pass

    info("Something went wrong decrypting " + str(title) + " URL: " + str(url))
    shared_state.values["database"]("sessions").delete("nx")
    return []



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337
#
# Special note: The signatures of all handlers must stay the same so we can neatly call them in download()
# Same is true for every get_xx_download_links() function in sources/xx.py

import json

from quasarr.downloads.linkcrypters.hide import decrypt_links_if_hide
from quasarr.downloads.sources.al import get_al_download_links
from quasarr.downloads.sources.by import get_by_download_links
from quasarr.downloads.sources.dd import get_dd_download_links
from quasarr.downloads.sources.dt import get_dt_download_links
from quasarr.downloads.sources.dw import get_dw_download_links
from quasarr.downloads.sources.mb import get_mb_download_links
from quasarr.downloads.sources.nx import get_nx_download_links
from quasarr.downloads.sources.sf import get_sf_download_links, resolve_sf_redirect
from quasarr.downloads.sources.sl import get_sl_download_links
from quasarr.downloads.sources.wd import get_wd_download_links
from quasarr.providers.log import info
from quasarr.providers.notifications import send_discord_message
from quasarr.providers.statistics import StatsHelper


def handle_unprotected(shared_state, title, password, package_id, imdb_id, url,
                       mirror=None, size_mb=None, links=None, func=None, label=""):
    if func:
        links = func(shared_state, url, mirror, title)

    if links:
        info(f"Decrypted {len(links)} download links for {title}")
        send_discord_message(shared_state, title=title, case="unprotected", imdb_id=imdb_id, source=url)
        added = shared_state.download_package(links, title, password, package_id)
        if not added:
            fail(title, package_id, shared_state,
                 reason=f'Failed to add {len(links)} links for "{title}" to linkgrabber')
            return {"success": False, "title": title}
    else:
        fail(title, package_id, shared_state,
             reason=f'Offline / no links found for "{title}" on {label} - "{url}"')
        return {"success": False, "title": title}

    StatsHelper(shared_state).increment_package_with_links(links)
    return {"success": True, "title": title}


def handle_protected(shared_state, title, password, package_id, imdb_id, url,
                     mirror=None, size_mb=None, func=None, label=""):
    links = func(shared_state, url, mirror, title)
    if links:
        valid_links = [pair for pair in links if "/404.html" not in pair[0]]

        # If none left, IP was banned
        if not valid_links:
            fail(
                title,
                package_id,
                shared_state,
                reason=f'IP was banned during download of "{title}" on {label} - "{url}"'
            )
            return {"success": False, "title": title}
        links = valid_links

        info(f'CAPTCHA-Solution required for "{title}" at: "{shared_state.values['external_address']}/captcha"')
        send_discord_message(shared_state, title=title, case="captcha", imdb_id=imdb_id, source=url)
        blob = json.dumps({"title": title, "links": links, "size_mb": size_mb, "password": password})
        shared_state.values["database"]("protected").update_store(package_id, blob)
    else:
        fail(title, package_id, shared_state,
             reason=f'No protected links found for "{title}" on {label} - "{url}"')
        return {"success": False, "title": title}
    return {"success": True, "title": title}


def handle_al(shared_state, title, password, package_id, imdb_id, url, mirror, size_mb):
    data = get_al_download_links(shared_state, url, mirror, title, password)
    links = data.get("links", [])
    title = data.get("title", title)
    password = data.get("password", "")
    return handle_unprotected(
        shared_state, title, password, package_id, imdb_id, url,
        links=links,
        label='AL'
    )


def handle_by(shared_state, title, password, package_id, imdb_id, url, mirror, size_mb):
    links = get_by_download_links(shared_state, url, mirror, title)
    if not links:
        fail(title, package_id, shared_state,
             reason=f'Offline / no links found for "{title}" on BY - "{url}"')
        return {"success": False, "title": title}

    decrypted = decrypt_links_if_hide(shared_state, links)
    if decrypted and decrypted.get("status") != "none":
        status = decrypted.get("status", "error")
        links = decrypted.get("results", [])
        if status == "success":
            return handle_unprotected(
                shared_state, title, password, package_id, imdb_id, url,
                links=links, label='BY'
            )
        else:
            fail(title, package_id, shared_state,
                 reason=f'Error decrypting hide.cx links for "{title}" on BY - "{url}"')
            return {"success": False, "title": title}

    return handle_protected(
        shared_state, title, password, package_id, imdb_id, url,
        mirror=mirror,
        size_mb=size_mb,
        func=lambda ss, u, m, t: links,
        label='BY'
    )


def handle_sf(shared_state, title, password, package_id, imdb_id, url, mirror, size_mb):
    if url.startswith(f"https://{shared_state.values['config']('Hostnames').get('sf')}/external"):
        url = resolve_sf_redirect(url, shared_state.values["user_agent"])
    elif url.startswith(f"https://{shared_state.values['config']('Hostnames').get('sf')}/"):
        data = get_sf_download_links(shared_state, url, mirror, title)
        url = data.get("real_url")
        if not imdb_id:
            imdb_id = data.get("imdb_id")

    if not url:
        fail(title, package_id, shared_state,
             reason=f'Failed to get download link from SF for "{title}" - "{url}"')
        return {"success": False, "title": title}

    return handle_protected(
        shared_state, title, password, package_id, imdb_id, url,
        mirror=mirror,
        size_mb=size_mb,
        func=lambda ss, u, m, t: [[url, "filecrypt"]],
        label='SF'
    )


def handle_sl(shared_state, title, password, package_id, imdb_id, url, mirror, size_mb):
    data = get_sl_download_links(shared_state, url, mirror, title)
    links = data.get("links")
    if not imdb_id:
        imdb_id = data.get("imdb_id")
    return handle_unprotected(
        shared_state, title, password, package_id, imdb_id, url,
        links=links,
        label='SL'
    )


def handle_wd(shared_state, title, password, package_id, imdb_id, url, mirror, size_mb):
    data = get_wd_download_links(shared_state, url, mirror, title)
    links = data.get("links", []) if data else []
    if not links:
        fail(title, package_id, shared_state,
             reason=f'Offline / no links found for "{title}" on WD - "{url}"')
        return {"success": False, "title": title}

    decrypted = decrypt_links_if_hide(shared_state, links)
    if decrypted and decrypted.get("status") != "none":
        status = decrypted.get("status", "error")
        links = decrypted.get("results", [])
        if status == "success":
            return handle_unprotected(
                shared_state, title, password, package_id, imdb_id, url,
                links=links, label='WD'
            )
        else:
            fail(title, package_id, shared_state,
                 reason=f'Error decrypting hide.cx links for "{title}" on WD - "{url}"')
            return {"success": False, "title": title}

    return handle_protected(
        shared_state, title, password, package_id, imdb_id, url,
        mirror=mirror,
        size_mb=size_mb,
        func=lambda ss, u, m, t: links,
        label='WD'
    )


def download(shared_state, request_from, title, url, mirror, size_mb, password, imdb_id=None):
    if "lazylibrarian" in request_from.lower():
        category = "docs"
    elif "radarr" in request_from.lower():
        category = "movies"
    else:
        category = "tv"

    package_id = f"Quasarr_{category}_{str(hash(title + url)).replace('-', '')}"

    if imdb_id is not None and imdb_id.lower() == "none":
        imdb_id = None

    config = shared_state.values["config"]("Hostnames")
    flags = {
        'AL': config.get("al"),
        'BY': config.get("by"),
        'DD': config.get("dd"),
        'DT': config.get("dt"),
        'DW': config.get("dw"),
        'MB': config.get("mb"),
        'NX': config.get("nx"),
        'SF': config.get("sf"),
        'SL': config.get("sl"),
        'WD': config.get("wd")
    }

    handlers = [
        (flags['AL'], handle_al),
        (flags['BY'], handle_by),
        (flags['DD'], lambda *a: handle_unprotected(*a, func=get_dd_download_links, label='DD')),
        (flags['DT'], lambda *a: handle_unprotected(*a, func=get_dt_download_links, label='DT')),
        (flags['DW'], lambda *a: handle_protected(*a, func=get_dw_download_links, label='DW')),
        (flags['MB'], lambda *a: handle_protected(*a, func=get_mb_download_links, label='MB')),
        (flags['NX'], lambda *a: handle_unprotected(*a, func=get_nx_download_links, label='NX')),
        (flags['SF'], handle_sf),
        (flags['SL'], handle_sl),
        (flags['WD'], handle_wd),
    ]

    for flag, fn in handlers:
        if flag and flag.lower() in url.lower():
            return {"package_id": package_id,
                    **fn(shared_state, title, password, package_id, imdb_id, url, mirror, size_mb)}

    if "filecrypt" in url.lower():
        return {"package_id": package_id, **handle_protected(
            shared_state, title, password, package_id, imdb_id, url, mirror, size_mb,
            func=lambda ss, u, m, t: [[u, "filecrypt"]],
            label='filecrypt'
        )}

    info(f'Could not parse URL for "{title}" - "{url}"')
    StatsHelper(shared_state).increment_failed_downloads()
    return {"success": False, "package_id": package_id, "title": title}


def fail(title, package_id, shared_state, reason="Offline / no links found"):
    try:
        info(f"Reason for failure: {reason}")
        StatsHelper(shared_state).increment_failed_downloads()
        blob = json.dumps({"title": title, "error": reason})
        stored = shared_state.get_db("failed").store(package_id, json.dumps(blob))
        if stored:
            info(f'Package "{title}" marked as failed!"')
            return True
        else:
            info(f'Failed to mark package "{title}" as failed!"')
            return False
    except Exception as e:
        info(f'Error marking package "{package_id}" as failed: {e}')
        return False



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import html
import time
from base64 import urlsafe_b64encode
from datetime import datetime, timezone

from quasarr.providers.imdb_metadata import get_localized_title
from quasarr.providers.log import info, debug
from quasarr.providers.sessions.dd import create_and_persist_session, retrieve_and_validate_session

hostname = "dd"
supported_mirrors = ["ironfiles", "rapidgator", "filefactory"]


def convert_to_rss_date(unix_timestamp):
    parsed_date = datetime.fromtimestamp(unix_timestamp, tz=timezone.utc)
    rss_date = parsed_date.strftime('%a, %d %b %Y %H:%M:%S %z')

    return rss_date


def extract_size(size_in_bytes):
    return {"size": size_in_bytes, "sizeunit": "B"}


def dd_feed(*args, **kwargs):
    return dd_search(*args, **kwargs)


def dd_search(shared_state, start_time, request_from, search_string="", mirror=None, season=None, episode=None):
    releases = []
    dd = shared_state.values["config"]("Hostnames").get(hostname.lower())
    password = dd

    if not "arr" in request_from.lower():
        debug(f'Skipping {request_from} search on "{hostname.upper()}" (unsupported media type)!')
        return releases

    dd_session = retrieve_and_validate_session(shared_state)
    if not dd_session:
        info(f"Could not retrieve valid session for {dd}")
        return releases

    if mirror and mirror not in supported_mirrors:
        debug(f'Mirror "{mirror}" not supported by "{hostname.upper()}". Supported mirrors: {supported_mirrors}.'
              ' Skipping search!')
        return releases

    imdb_id = shared_state.is_imdb_id(search_string)
    if imdb_id:
        search_string = get_localized_title(shared_state, imdb_id, 'en')
        if not search_string:
            info(f"Could not extract title from IMDb-ID {imdb_id}")
            return releases
        search_string = html.unescape(search_string)

    qualities = [
        "disk-480p",
        "web-480p",
        "movie-480p-x265",
        "disk-1080p-x265",
        "web-1080p",
        "web-1080p-x265",
        "web-2160p-x265-hdr",
        "movie-1080p-x265",
        "movie-2160p-webdl-x265-hdr"
    ]

    headers = {
        'User-Agent': shared_state.values["user_agent"],
    }

    try:
        release_list = []
        for page in range(0, 100, 20):
            url = f'https://{dd}/index/search/keyword/{search_string}/qualities/{','.join(qualities)}/from/{page}/search'

            releases_on_page = dd_session.get(url, headers=headers, timeout=10).json()
            if releases_on_page:
                release_list.extend(releases_on_page)

        for release in release_list:
            try:
                if release.get("fake"):
                    debug(
                        f"Release {release.get('release')} marked as fake. Invalidating {hostname.upper()} session...")
                    create_and_persist_session(shared_state)
                    return []
                else:
                    title = release.get("release")

                    if not shared_state.is_valid_release(title,
                                                         request_from,
                                                         search_string,
                                                         season,
                                                         episode):
                        continue

                    imdb_id = release.get("imdbid", None)

                    source = f"https://{dd}/"
                    size_item = extract_size(release.get("size"))
                    mb = shared_state.convert_to_mb(size_item) * 1024 * 1024
                    published = convert_to_rss_date(release.get("when"))
                    payload = urlsafe_b64encode(
                        f"{title}|{source}|{mirror}|{mb}|{password}|{imdb_id}".encode("utf-8")).decode("utf-8")
                    link = f"{shared_state.values['internal_address']}/download/?payload={payload}"

                    releases.append({
                        "details": {
                            "title": title,
                            "hostname": hostname.lower(),
                            "imdb_id": imdb_id,
                            "link": link,
                            "mirror": mirror,
                            "size": mb,
                            "date": published,
                            "source": source
                        },
                        "type": "protected"
                    })
            except Exception as e:
                info(f"Error parsing {hostname.upper()} feed: {e}")
                continue

    except Exception as e:
        info(f"Error loading {hostname.upper()} feed: {e}")

    elapsed_time = time.time() - start_time
    debug(f"Time taken: {elapsed_time:.2f}s ({hostname})")

    return releases



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import setuptools

from quasarr.providers.version import get_version

try:
    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()
except:
    import io

    long_description = io.open('README.md', encoding='utf-8').read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="quasarr",
    version=get_version(),
    author="rix1337",
    author_email="",
    description="Quasarr connects JDownloader with Radarr, Sonarr and LazyLibrarian. It also decrypts links protected by CAPTCHAs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rix1337/Quasarr",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=required,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'quasarr = quasarr:run',
        ],
    },
)
